#!/usr/bin/env python2.1

##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################

import os,sys,re,getopt
import urlparse,time,shutil
import BaseHTTPServer
from types import FileType,StringType
from websiteloadtest.common import myhttplib
from websiteloadtest.common.Request import Request

__version__     = "$Id: proxy.py,v 1.4 2002/03/20 13:57:34 andreas Exp $"

# hostname 
localHost       = ''

# our local port
localPort       = 2222

# outputfile for requests
outfile         = 'replay.dat'

# File handle or filename for debugging
debug_fp        = sys.stderr

# additional debugging
debug_level     = 0

# Record only POST requests
postonly        = 0

# Directory for POST data
postDataDir     = 'proxy_postdata'

# Directory for retrieved objects
proxyDataDir     = 'proxy_data'

# shall we store retrieved objects
storeobjects    = 0 

interesting_headers = None # by default, keep all of them


############################################################################
# Don't change anything below
############################################################################

# regex to extract content-length from header
reg_cl = re.compile("content-length: (\w)",re.I)

# timestamp start of test
start_ts = time.time()


numProxyData = numPosts = 0

# Create storage directories
for d in (postDataDir,proxyDataDir):
    if not os.path.exists(d):
        os.makedirs(d)

# determine the last used number for post data
files = os.listdir(postDataDir)
files = filter(lambda x: x.startswith('post.'),files)
files.sort()
try:
    numPosts = int(files[-1].split('.')[-1]) + 1
except: pass

# determine the last used number for proxy data
files = os.listdir(proxyDataDir)
files = filter(lambda x: x.startswith('proxy.'),files)
files.sort()
try:
    numProxyData = int(files[-1].split('.')[-1]) + 1
except: pass


def debug(*args):
    """ write debug to file or file descriptor """

    if debug_level==0: return

    if isinstance(debug_fp,FileType) and not debug_fp.closed:
        for a in args: debug_fp.write(str(a)+' ')
        debug_fp.write('\n')

    elif isinstance(debug_fp,StringType):
        fp = open(debug_fp,'a')
        for a in args: fp.write(str(a)+' ')
        fp.write('\n')
        fp.close()
    
    else:
        raise RuntimeError, "Unsupported output channel"   


def init():
    shutil.rmtree(proxyDataDir)
    shutil.rmtree(postDataDir)
    os.mkdir(postDataDir)
    os.mkdir(proxyDataDir)
    if os.path.exists(outfile): os.unlink(outfile)


class ReqHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """ our proxy server """

    def log(self,r):
        fp = open(outfile,'a')
        fp.write(str(r) + '\n')
        fp.flush()
        fp.close()


    def do_XXX(self):
        global numPosts, numProxyData

        addHeaders = {}

        # retrieve the headers from the request

        for line in self.headers.headers:
            fields = line.strip().split(':')
            k = fields[0].lower()
            v = ':'.join( fields[1:])
            if interesting_headers is None or k in interesting_headers:
                addHeaders[k.lower()] = v


        # determine relevant data for POST request

        if self.command == 'POST':
            length = self.headers.getheader('content-length','0')
            postdata = self.rfile.read(int(length))

            postdatafile = os.path.join(postDataDir,'post.%05d' % numPosts)
            numPosts+=1
            open(postdatafile,'w').write(postdata)

            self.rfile.close()
        else:
            postdata = ''


        # determine path info

        tp = urlparse.urlparse(self.path)
        p  = self.path.replace("%s://%s" % (tp[0],tp[1]),'')

        # send request to other server
        errcode,errmsg,headers,data = self.send_srv(
                        self.command,
                        p,
                        self.headers,
                        postdata
                        )


        # send response back to client        
        self.send_response(errcode,errmsg)

        # send HTTP headers

        for hd in headers.headers:
            fields = filter(lambda x: x!='', hd.strip().split(":"))
            if fields==[]: continue

            k = fields[0]
            v = ':'.join(fields[1:])

            self.send_header(k,v)

        self.end_headers()

        # send the received body

        self.wfile.write(data)

        # store received data in file

        outputfile = None

        if storeobjects:
            outputfile = os.path.join(proxyDataDir,'proxy.%08d' % numProxyData)
            fp = open(outputfile,'w')
            fp.write(data)
            fp.close()
            numProxyData+=1


        # Create request object to be written to file

        req = Request(  
                    timestamp = time.time()-start_ts,
                    path=tp[2],
                    params=tp[3],
                    query=tp[4],
                    fragment=tp[5],
                    method=self.command,
                    cond_get=0,
                    headers=addHeaders,
                    outputfile=outputfile,
                    )                                                

        if self.command=="POST":
            req.setPostData( "PostDataFile('%s')" % postdatafile)

        if postonly==0 or (postonly==1 and self.command=="POST"):        
            self.log(req)


    do_HEAD = do_GET = do_POST = do_XXX

    def send_srv(self, method,path, headers='', body=''):

        x = headers.getheader('host')
        if x.find(':')>-1:
            remoteHost,remotePort = x.split(':')
        else:
            remoteHost,remotePort = x,80
        

        h = myhttplib.HTTP(remoteHost,int(remotePort))
        h.putrequest(method,path)

        debug("sending -----------------------------------")
        debug("Remote: %s:%s" % (remoteHost,remotePort))
        debug(method,path)
        debug(headers)

        for hd in headers.headers:
            fields = filter(lambda x: x!='',hd.strip().split(":"))
            if len(fields)==0: continue
        
            k = fields[0]
            v = ':'.join(fields[1:])

            # Bad hack !!!
            if k.lower() in ['content-length'] and method=="POST":
                v = str(int(v) + 1)

            h.putheader(k,v)

        h.endheaders()

        if method == 'POST': 

            debug('sending post body:')
            debug(body)

            # Bad hack - usually we don't need to add \n - but for Zope 2.3 !
            h.send(body+'\n')

        errcode,errmsg,headers = h.getreply()

        debug("receiving -----------------------------------")
        debug(headers)

        cl = headers.getheader('content-length',None)

        if cl:      data = h.getfile().read(int(cl))
        else:       data = h.getfile().read() 

        debug("Data:")
        debug(data)

        return errcode,errmsg,headers,data


def usage():
    print "%s [options] " % os.path.basename(sys.argv[0])
    print
    print "Options:"
    print "    -h|--help           This help"
    print "    -d|--debug"
    print "    -i|--init           remove existing data"
    print "    --postonly          record only POST requests"
    print "    --storeobjects      store retrieved objects to disk"
    print "    -p|--port <PORT>    local port to listen to"
    print "    -m|--host <HOST>    hostname/IP to listen to"
    print "    -o|--output <FILE>  use a custom output file"
    print
    print "Version:",__version__


def main():
    global localHost,localPort,debug_level,debug_fp,postonly,storeobjects

    opts,args = getopt.getopt(sys.argv[1:],
                              'ihdo:p:m:',
                              ['port=',
                               'help',
                               'debug',
                               'output=',
                               'postonly',
                               'host=',
                               'storeobjects',
                               'init',
                               'interesting-headers='
                              ])

    for k,v in opts:
        if k in ['--help','-h']:            usage(); sys.exit(1)
        if k in ['--debug','-d']:           debug_level = 1
        if k in ['--port','-p']:            localPort   = int(v)
        if k in ['--host','-m']:            localHost   = v
        if k in ['--output','-o']:          debug_fp    = v
        if k in ['--postonly']:             postonly    = 1
        if k in ['--storeobjects']:         storeobjects = 1
        if k in ['--init','-i']:            init()

        if k in ['--interesting-headers']:
            global interesting_headers
            interesting_headers = [ x.strip().lower() for x in v.split(',') ]

    print "Proxy listening on host: %s, port: %d" % (localHost or "all IPs",localPort)

    proxy = BaseHTTPServer.HTTPServer( (localHost,localPort), ReqHandler)
    proxy.serve_forever()

    sys.exit(0)

if __name__=='__main__':
    main()
