##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################

import threading
from websiteloadtest.common.xmlrpcreply import XRR

class Client:

    def __init__(self,host,port,cookie,alias):
        self.host   = host
        self.port   = port
        self.cookie = cookie
        self.active = 1
        self.tries  = 0

    def setActive(self):
        self.active = 1
        self.tries  = 0

    def setInactive(self):
        self.active = 0

    def incTries(self):
        self.tries+=1
        return self.tries

    def asDict(self):
        return { 'host': self.host,
                 'port': self.port,
                 'active': self.active
                }


class Registry:
    """ registry for clients """
    
    clients = { }


    def __init__(self):
        self._lock = threading.Lock()

    def acquire(self): self._lock.acquire()
    def release(self): self._lock.release()


    def heartbeat(self,hostname=None):
        """ heartbeat method for client requests """

        if self.has_key(hostname):
        
            self.acquire()
            self[hostname].setActive()
            self.release()

            return XRR(1,'ok')            

        else:
            
            return XRR(0,'need to re-register')


    def __getitem__(self,host):  return self.clients[host]
    def __setitem__(self,k,v):   self.clients[k] = v
    def has_key(self,key):       return self.clients.has_key(key)
    def __len__(self):           return len(self.clients)


    def register_client(self,host,port,cookie,alias ):

        self.acquire()
        self[alias] = Client(host,port,cookie,alias)
        self.release()

        return 1


    def unregister_client(self,host,cookie=''):


        if self.has_key(host):
            self.acquire()
            del self.clients[host] 
            self.release()

            return 1
        else:
            return 0


    def getClientInfo(self,host):
        return self[host]

    def allClients(self):
        """ return a list of names of all registered clients """

        lst = self.clients.keys()
        lst.sort()

        return lst
